@@ -4,47 +4,41 @@ $^W = 1;
 use Socket qw ( inet_aton );
 
 use vars qw ( $VERSION );
-$VERSION = '212.005'; # DEC 2002, version 0.04
+$VERSION = '212.006'; # DEC 2002, version 0.06
 
 my $singleton = undef;
-my %ip_db;
+my $ip_db;
+my $null = substr(pack('N',0),0,1);
 my %cc;
 my $tld_match = qr/\.([a-zA-Z][a-zA-Z])$/o;
 my $ip_match = qr/^(\d|[01]?\d\d|2[0-4]\d|25[0-5])\.(\d|[01]?\d\d|2[0-4]\d|25[0-5])\.(\d|[01]?\d\d|2[0-4]\d|25[0-5])\.(\d|[01]?\d\d|2[0-4]\d|25[0-5])$/o;
 
-my @dton;
+my $bit0 = substr(pack('N',2 ** 31),0,1);
+my $bit1 = substr(pack('N',2 ** 30),0,1);
+my @mask;
 my @dtoc;
 {
-    for (my $i = 0; $i <= 31; $i++){
-	$dton[$i] = pack('N', 2**32 - 2**$i);
+    for (my $i = 0; $i <= 32; $i++){
+	$mask[$i] = pack('N',2 ** (32 - $i));
     }
+
     for (my $i = 0; $i <= 255; $i++){
 	$dtoc[$i] = substr(pack('N',$i),3,1);
     }
     (my $module_dir = $INC{'IP/Country/Fast.pm'}) =~ s/\.pm$//;
 
+    local $/;   # set it so <> reads all the file at once
+
     open (IP, "< $module_dir/ip.gif")
-	or die ("couldn't create IP database: $!");
+	or die ("couldn't read IP database: $!");
     binmode IP;
-    my $ip_ultra;
-    {
-	local $/;   # set it so <> reads all the file at once
-	$ip_ultra = <IP>;  # read in the file
-    }
+    $ip_db = <IP>;
     close IP;
-    my $ip_num = (length $ip_ultra) / 6;
-    for (my $i = 0; $i < $ip_num; $i++){
-	$ip_db{substr($ip_ultra,6 * $i,5)} = substr($ip_ultra,6 * $i + 5,1);
-    }
 
     open (CC, "< $module_dir/cc.gif")
-	or die ("couldn't create country database: $!");
+	or die ("couldn't read country database: $!");
     binmode CC;
-    my $cc_ultra;
-    {
-	local $/;   # set it so <> reads all the file at once
-	$cc_ultra = <CC>;  # read in the file
-    }
+    my $cc_ultra = <CC>;  # read in the file
     close CC;
     my $cc_num = (length $cc_ultra) / 3;
     for (my $i = 0; $i < $cc_num; $i++){
@@ -79,14 +73,30 @@ sub inet_atocc ($)
 sub inet_ntocc ($)
 {
     my $inet_n = $_[1] || $_[0];
-    for (my $i = 31; $i >= 0; $i--)
-    {
-	my $masked_ip = $inet_n & $dton[$i];
-	my $key = $masked_ip.$dtoc[$i];
-	if (exists $ip_db{$key}){
-	    return $cc{$ip_db{$key}};
+
+    my $pos = 0;
+#    print STDERR unpack('B32',$inet_n)."\n";
+#    print STDERR "position: $pos\n";
+    for (my $i = 1; $i <= 32; $i++){
+	my $byte_zero = substr($ip_db,$pos,1);
+	if (($byte_zero & $bit0) eq $bit0){ # country code
+	    if (($byte_zero & $bit1) eq $bit1){ # unpopular country code - skip a byte
+		return $cc{substr($ip_db,$pos+1,1)};
+	    } else { # popular country code
+		return $cc{$byte_zero ^ $bit0};
+	    }
+	} else {
+	    if (($inet_n & $mask[$i]) eq $mask[$i]){
+		my $jump = unpack('N',$null.substr($ip_db,$pos,3));
+		$pos = $pos + 3 + $jump;
+#		print STDERR "ONE: jumping: $jump\n";
+	    } else {
+		$pos = $pos + 3;
+#		print STDERR "ZERO\n";
+	    }
 	}
     }
+#    print STDERR "\n";
     return undef;
 }
 
